#include "fitness.h"


class KeywordsMeanXover : public eoQuadOp<Keywords> {

public:
   
  string className() const { return "KeywordsMeanCrossover"; }
  
  bool operator()(Keywords& ind1, Keywords& ind2) {
    
    if ( ind1.size() != ind2.size() ) {
      // not implemented yet
      return false; 
    }  

    // match centers and create new in between
    offspring1.resize(ind1.size()); 
    offspring2.resize(ind2.size());

    int index; double r; 
    for (unsigned i=0; i<ind1.size(); i++) {
      // find nearest to ind1[i] 
      index = nearest(ind1[i],ind2); 
      
      offspring1[i].mean(ind1[i],ind2[index],r); 
      offspring2[i].mean(ind1[i],ind2[index],1-r);
      
      ind2.remove(index);
    } 
    
    ind1 = offspring1;
    ind2 = offspring2; 

    return true;
  } 

  int nearest(Vector& x,Keywords& ind) {
    int ret = 0;
    double dist = x.distance(ind[0]); 
    double d; 

    for (unsigned i=1; i<ind.size(); i++) 
      if ( (d = x.distance(ind[i])) < dist) {
	dist  = d;
	ret = i;
      } 
    return ret; 
  } 


private:
  Keywords offspring1;
  Keywords offspring2; 

}; 

class KeywordsXover : public eoQuadOp<Keywords> {

public:  
  string className() const { return "KeywordsCrossover"; }

  bool operator()(Keywords& ind1, Keywords & ind2) 
  {
    //    assert( ind1.size() == ind2.size() );
    int minsize = (ind1.size() < ind2.size()) ? ind1.size() : ind2.size(); 

    int crosspoint1 = rng.random( minsize ); 
   

    for (int i=0; i<crosspoint1; i++) {
      Vector buff = ind1[i];
      ind1[i] = ind2[i];
      ind2[i] = buff;
    }
    
    return true; // vraci true, pokud alespon jeden z jedincu byl zmenen 
  }
};

class KeywordsUniformXover : public eoQuadOp<Keywords> {

public:  
  string className() const { return "UniformCrossover"; }

  bool operator()(Keywords& ind1, Keywords & ind2) 
  {

    for (unsigned i=0; i<ind1.size(); i++) {
      if (rng.flip()) {
	Vector buff = ind1[i];
	ind1[i] = ind2[i];
	ind2[i] = buff;
      }
    }
    
    return true; // vraci true, pokud alespon jeden z jedincu byl zmenen 
  }
};


class KeywordsRemove : public eoMonOp<Keywords> 
{
public:
  
  string className() const { return "KeywordsRemove"; }

  bool operator()(Keywords& ind) {
    if (ind.size()<=2) 
      return false; 
    
    /*
    // find smallest cluster and remove it
    int min = 0; 
    for (int i=1; i<ind.size(); i++) 
      if (ind.counters[i] < ind.counters[min]) 
	min = i;
    ind.remove(min); 
    */ 

    ind.remove( rng.random( ind.size() ) );
    return true; 
  }; 
}; 

class KeywordsAdd : public eoMonOp<Keywords> {
public:

  KeywordsAdd(DataSet& _data) : data(_data) {};
  
  string className() const { return "KeywordsAdd"; }

  bool operator()(Keywords& ind) {

    // add random cluster 
    Vector& x  = data[ rng.random( data.size() ) ].input; 
    ind.add( x ); 

    /*
    if (rng.flip()) {
      Vector& y  = data[ rng.random( data.size() ) ].input; 
      ind.add( y ); 
      } */
    
    /*
    // find the biggest cluster and split it 
    int max = 0; 
    for (int i=1; i<ind.size(); i++) 
      if (ind.counters[i] > ind.counters[max]) 
	max = i;
    Vector x;
    x = ind[max]; 
    for (unsigned j=0; j<x.size(); j++) 
      x[j] += rng.uniform(0.0001);        // should be data dependent (much smaller than cluster size) 
    ind.add( x );
    return true; 
    */
  };

private:
  DataSet& data; 
}; 


class KeywordsMut1: public eoMonOp<Keywords>
{
public:
  
  KeywordsMut1(const vector<double>& _min,const vector<double>& _max) : min(_min), max(_max) {
    prob = 0.2; 
  };

  string className() const { return "KeywordsMut1"; }

  
  bool operator()(Keywords & ind) {
       

    for (unsigned change=0; change<ind.size(); change++) 
      if (rng.uniform()<prob) 
	for (unsigned i=0; i<ind[change].size(); i++) {
	  ind[change][i] = min[i] +  rng.uniform(max[i]-min[i]);
	} 

    return true; // vraci true, pokud byl jedinec  zmenen 
  }

private:
    double prob;   // mutation per center probability 
    vector<double> min;
    vector<double> max; 
};

class KeywordsMut2: public eoMonOp<Keywords>
{
public:
  
  KeywordsMut2(const vector<double>& _min,const vector<double>& _max) : min(_min), max(_max) {

    interval.resize( min.size() );
    for (unsigned i=0; i<interval.size(); i++) {
      interval[i] = max[i] - min[i];
      interval[i] /= 20.0;
    }

    prob = 0.2; 
  };

  string className() const { return "KeywordsMut2"; }

  
  bool operator()(Keywords & ind) {
       

    for (unsigned change=0; change<ind.size(); change++) 
      if (rng.uniform()<prob) 
	for (unsigned i=0; i<ind[change].size(); i++) {
	  ind[change][i] += -interval[i] + rng.uniform(2*interval[i]);
	} 

    return true; // vraci true, pokud byl jedinec  zmenen 
  }

private:
    double prob;   // mutation per center probability 
    vector<double> min;
    vector<double> max; 
    vector<double> interval; 
};


class KMeansMut : public eoMonOp<Keywords>
{
public:
  
  KMeansMut(DataSet& _data) : data(_data) {  
    maxIter = 5; 
  };
  

  string className() const { return "KMeansMutation"; }

  
  // max maxIter iterations of k-means algorithm
  bool operator()(Keywords & ind) {
    
    // one step of k-means

    // setup 
    counters.resize( ind.size() );
    centers.resize( ind.size() );

    for (int iter=0; iter<maxIter; iter++) {

      for (unsigned i=0;  i<counters.size(); i++) counters[i] = 0; 
      for (unsigned i=0;  i<centers.size(); i++) centers[i].zero( data.getN() );
    
      // go through training set 
      int index;
      for (int i=0; i<data.size(); i++) {
	index = ind.nearest(data[i].input);
	counters[index]++;
	centers[index] += data[i].input;
      }
     
      // comp means and move centers
      for (unsigned i=0; i<centers.size(); i++) 
	if (counters[i] > 0) 
	  centers[i] /= counters[i];
	else 
	  centers[i] = data[ rng.random(data.size()) ].input; 
	  
      if (ind.centers == centers) break; // !porovnavani realnych vektoru (vzdalenost lepsi, ale prijde mi 
                                         // zbytecna)
      ind.centers = centers;
      iter++; 
    }

    return true; // vraci true, pokud byl jedinec  zmenen 
  }

private:
  int maxIter; 
  vector<int> counters;
  vector<Vector> centers;
  DataSet& data; 

};
